#include <stdio.h>
#include <strings.h>
#include <unistd.h>
#include <errno.h>
#include <termios.h>
#include <fcntl.h>
#include <time.h>
#include <stdlib.h>

#define STATUS_SIZE 16

#if 0
#define HIGH_SPEED_RS232
#endif

#ifdef HIGH_SPEED_RS232
#define RS232_SPEED 115200
#else
#define RS232_SPEED 9600
#endif

static int newsfd;
static struct termios oldtio;
static char commDevice[]="/dev/cuaa0";
 
static int InitComm(int initBaudRate) {
  struct termios newtio;

  /* open the device to be non-blocking (read will return immediately) */
  newsfd = open(commDevice, O_RDWR | O_NOCTTY | O_NONBLOCK);
  if (newsfd <0) {perror(commDevice); exit(-1); }
  
  /* allow the process to receive SIGIO */
  fcntl(newsfd, F_SETOWN, getpid());
  /* Make the file descriptor asynchronous (the manual page says only
     O_APPEND and O_NONBLOCK, will work with F_SETFL...) */
  fcntl(newsfd, F_SETFL, O_ASYNC);
  
  tcgetattr(newsfd,&oldtio); /* save current port settings */
  /* set new port settings for canonical input processing */
  bzero(&newtio,sizeof(newtio));
  newtio.c_cflag = CS8 | CLOCAL | CREAD ; 
  newtio.c_iflag = IGNPAR;
  newtio.c_oflag = 0;
  newtio.c_lflag = 0;
  newtio.c_cc[VMIN] = 0;
  newtio.c_cc[VTIME] = 0;
  newtio.c_ispeed = initBaudRate;
  newtio.c_ospeed = initBaudRate;
  tcflush(newsfd, TCIFLUSH);
  if (tcsetattr(newsfd,TCSANOW,&newtio)==-1) {
    printf("Serial port initialization failed, errno: %d\n",errno);
    return -1;
  }

  return 0;
}

static int SendChar(char d) {
  if (write(newsfd, &d, 1)<0)
    return -1;
  else
    return d;
}


#if 0
static int SendData(char *data, int len) {
  if (write(newsfd, data, len)<0) 
    printf("Write error\n"); 
  return 0;
}
#endif

static int ReceiveChar(void) {
    unsigned char d;
  if (read(newsfd, &d, 1) == 1) {
    return d & 0xff;
  } else {
    return -1;
  }
}

static int ReceiveData(char *data, int len) { 
  int i = 0;
  int startTime = time(NULL);

  while(i<len && time(NULL) - startTime < 2) { 
    if(read(newsfd,data+i,1)==1)
      i++;       
  }
  return i;
}

#if 0
static int CloseComm(void)  { 
  /* restore old port settings */
  tcsetattr(newsfd,TCSANOW,&oldtio);
  /* close serial port */
  close(newsfd); 
  return 0; 
}
#endif


int main(int argc, char **argv) {
  int i, c;

  if (InitComm(RS232_SPEED))
    exit(-1);

  while(1) {
      char d;

      if ((c = ReceiveChar()) >= 0) {
	  fputc(c, stdout);
      }
  }

  return 0;
}
