#ifndef MIDI_H
#define MIDI_H
#include <vstypes.h>
#ifndef ASM
#include <stdio.h>
#endif

//#include "sin.h"
#define SHORTSIN 2048 /*2048 or 1024 or not defined */

#define WAVE_MEM_IN_Y
#ifdef WAVE_MEM_IN_Y
#define WAVE_MEM __y
#else
#define WAVE_MEM __x
#endif

#ifdef __VSDSP__
auto s_int16 WmaStereoCopy(s_int16 *p, s_int16 samples);
auto s_int16 AudioBufFill(void);
//#define AUDIO_BUFFER_SZ 4096
extern const s_int16 audioBufferSize;
#endif


#define FS 44100
#define FSPERMILLION (FS/1000000.0)
#define T  (1.0/FS)

#define DECODE_BUFFERS 16 /*Buffers to decode at once - 
			   This only affects the responce time and accuracy 
			   of external interrupts like skip and seek */
#define PI 3.14159265358979
#define TWOPI 6.2832
#define MAX_NOTE_POLYPHONY 40
// Try to keep room for at least 24 new notes/64 samples
#define READY_FOR_NOTES 24
#define NUMNOTES (MAX_NOTE_POLYPHONY+READY_FOR_NOTES)

#define MAXNOTE 16 //6
#define OSCBUFFER 64
#define BUFFER_SHIFT 6
#define ENV_STEP 16
#define STEP_SHIFT 4

//#define SHOWINIT
#define USE_GENERIC /* If CodMidiGeneric() is used */


#define CHANNELS 16
#define EOT 0
#define CONTINUE 1
#define ONEPERMAXVELOCITY 0.000062
/* 0.0010 means remove when 32..-31
   0.0015 means remove when 49..-15
   0.0020 means remove when 64..0
   0.0025 means remove when 81..18
*/
#define REMOVE_SILENT 0.002


#define BLOWNOISE
#define BANDPASS
#define REVERB
#define INIT_REVERB 1
#define DO_NOT_INIT_REVERB 0
#define REVERB_MIX 0.66
#define REVERB_FEEDBACK 0.15
#define REV_BUFFER 16
#define DECORDER 32 /* decimation/interpolation filter order 32, 20, or 16 */

//#define CHORUS
#ifdef CHORUS
#define INIT_CHORUS 1
#define INIT_CHORUS_SIN 2
#define DO_NOT_INIT_CHORUS 0
#define CHORUS_LENGTH 1024
#endif/*CHORUS*/

#define GLOBAL_PHASES 5
#define MAKE_ID(a,b,c,d) (((u_int32)(a)<<24)|((u_int32)(b)<<16)|((c)<<8)|(d))

#ifndef ASM
#ifdef __VSDSP__
#define FLOAT __fract int
#define MAXFRACT (FLOAT)0x7fff
#define MINFRACT (FLOAT)0x8000
#else
#ifndef __a
#define __a
#endif
#ifndef __c
#define __c
#endif
#define FLOAT float
#define __y
#endif

#define TREMOLO_LFO_FREQ (FLOAT)(64.0 * 2.0 * 12.0/FS)
#define EL_PIANO1_LFO_FREQ (FLOAT)(64.0 * 2.0 * 5.0/FS)
#define EL_PIANO2_LFO_FREQ (FLOAT)(64.0 * 2.0 * 4.0/FS)

#define FLOAT_ONE (FLOAT)(1/32768.0)
#define FLOAT_TWO (FLOAT)(2/32768.0)
#define FLOAT_THREE (FLOAT)(3/32768.0)


/* should really be in reverb.h */
typedef struct {
  __y FLOAT *db;
  __y FLOAT *dbEnd;
  __y FLOAT *ap;
  __y FLOAT *apEnd;

  FLOAT b;
  FLOAT g;
  FLOAT out;
  FLOAT flt;

  __y FLOAT *comb;
  __y FLOAT *allp;
} CombFilter;

typedef union {
  FLOAT phase;
  FLOAT freq;
  FLOAT gain;
  FLOAT attack;
  FLOAT decay;
  FLOAT sustain;
  FLOAT release;
  FLOAT filt;
  FLOAT env;
  s_int16 cutoff;
  s_int16 seed;
  s_int16 counter;
} PhaseUnion;

/**/
void CodMidiReverb( FLOAT *input, FLOAT gain );
void CodMidiReverbInit(void);
void CodMidiReverbSetDecay(int RT60);
void CodMidiInitComb(PhaseUnion *seed, register FLOAT f);
int CodMidiEgetc();
s_int32 CodMidiParse();

#else  /*ASM*/
/* defines for codMidiReverbs.s */
#define CF_DB 0
#define CF_DBEND 1
#define CF_AP 2
#define CF_APEND 3
#define CF_B 4
#define CF_G 5
#define CF_OUT 6
#define CF_FLT 7
#define CF_COMB 8
#define CF_ALLP 9
#define CF_SIZE 10
#endif /*ASM*/

#define LOWKEY 21
#define HIGHKEY 108
#define PITCHBENDSENSITIVITY 2
#define MODULATIONDEPTHRANGE 0.5
#define FBUFF 6 //filter intermediate buffer
#define PHASEMAX 14
#define INIT_INSTRUMENT 1
#define DONT_INIT_INSTRUMENT 0

#define CONGA_RELEASE_TIME 6000
#define CLAVES_RELEASE_TIME 1800

#ifndef ASM
extern void CodMidiInitControllers();
extern s_int32 CodMidiReadVariNum();
extern void CodMidiSetMIPTable();
extern void CodMidiSample();
extern u_int16 CodMidiRead16bit();
extern u_int32 CodMidiRead32bit();
extern void CodMidiInitTrack();
void CodMidiChannelMessage(int status,int c1,int c2);


#ifdef __VSDSP__
#ifndef STANDALONE
#ifdef SHORTSIN
#if SHORTSIN == 1024
extern const FLOAT __y shortSin1k[];
#else
extern const FLOAT __y shortSin2k[];
#endif
#else
extern const FLOAT __y lpcsin[];
#endif
#endif
#define WRAP(a) (a)
#define WRAPINT(a) (a)
#define FLOAT2INT(a) ((int)((FLOAT)(a)))
#define INT2FLOAT(a) ((FLOAT)(a))
#else
extern int verbose;
#ifdef SHORTSIN
#if SHORTSIN == 1024
extern const float shortSin1k[1024];
#else
extern const float shortSin2k[2048];
#endif
#else
extern const float lpcsin[4096];
#endif
extern float WRAP(float l);
extern int WRAPINT(int l);
//extern int FLOAT2INT(float l);
//extern float INT2FLOAT(int l);
#define FLOAT2INT(l) ((int)((l)*32768.0))
#define INT2FLOAT(l) ((int)(l)/32768.0)
#endif


typedef struct {
  FLOAT left;
  FLOAT right;
} Stereo;

typedef struct MidiNote MidiNote;
typedef void (*INSTRPTR)(MidiNote*,int);

/* DO NOT CHANGE THE STRUCTURE!
   SOME ASM FUNCTIONS DEPEND ON THE CURRENT OFFSETS!! */
struct MidiNote {
    struct MidiNote *nextNote;

    FLOAT prevAmpL;
    FLOAT prevAmpR;
    FLOAT prevEnv;  		/*Previous envelope gain*/
    s_int32 relTime;    	/*Note-off time in samples OR fade flag (==1)*/
    FLOAT prevEnv2;
    FLOAT prevEnv3;
    FLOAT filtEnv;		/*Previous filter envelope gain*/
    /* ---------------------------------------------------- */
    /* Do not change the sizes or order of the above fields */
    /* ---------------------------------------------------- */
    struct {
	unsigned int time:1;		/*Decay active*/
	unsigned int time2:1;
	unsigned int time3:1;
	unsigned int filt:1;
    } dec;
    s_int32 onTime;     		/*Note-on time in samples*/

    FLOAT f;			/*Frequency from freq table*/
    s_int16 pitch;			/*MIDI key number 21...108*/
    FLOAT filterBuffer[FBUFF];	/*Intermediate of lowpass filter (v1)*/
  PhaseUnion u[PHASEMAX];
    s_int16 baseCutoff;		/*Fractional table index*/
    s_int16 pc;
    FLOAT velocity;
    INSTRPTR instrument;     	/*Program change - instrument to play*/
};

typedef struct {
    s_int16 channel;		/*Channel number (0 = ch 1, 1 = ch 2, etc)*/
    s_int16 polyphony;		/*Cumulative polyphony*/
} MIP;

typedef struct MidiChannel {
    INSTRPTR instrument;	/*Program change - instrument to play*/
    s_int16 pc;			/*Program change number (original) */
    struct MidiNote *firstNote;
    FLOAT channelVolume;	/*Channel volume*/
    FLOAT expression;		/*Volume change*/
    FLOAT reverb;        /*Reverb gain 0...0.2*/
#ifdef CHORUS
    FLOAT chorus;        /*Chorus gain */
#endif/*CHORUS*/
    s_int16 pan;			/*Panning*/
    struct {
	u_int16 hold:1;		/*Hold (Damper) on/off*/
	u_int16 sustenuto:1;
    } u;
    s_int16 chPoly;			/*Channel polyphony*/
#define RPN  0
#define NRPN 1
  s_int16 RpnSel; /* 0 for Registered, 1 for NonRegistered */
  s_int16 Rpn[2];
  s_int16 bank;			/* Bank select for the next program change */
  s_int16 origBank;
  s_int16 pressure;
  s_int16 bendRange;		/*Maximum pitch bend in semitones up/down*/
  s_int16 pitchBend;		/*Pitch bend value, 0 == middle*/
  s_int16 modWheel;		/*Modulation depth of the channel*/
  s_int16 masterTune;
} MidiChannel;

#ifndef __VSDSP__
extern FILE *OUT;	/*PCM sample text file handle*/
#endif

s_int16 MidiUsesReverb(void);
void MidiMeta(u_int16 meta);
#if 1
int  CodMidiMidiInit(void);
void CodMidiMidiDecode(int ntracks);
#endif


#if 0
struct cymbalData {
    s_int16 cymbalIdx;
    s_int16 cymbalPhase;
    FLOAT cymbalSave[32];
};
#else

#define MAX_INTERP_SAMPLES 12
struct SampleInterp {
    FLOAT f;
    u_int32 phase;
    s_int16 bufIdx;
    s_int16 imaLeft;
    const u_int16 *sampleIma;
    short val;     /* last sample value */
    short state;   /* current state */
    FLOAT buf[3*32]; /* allow upto 2 blocks decoded per 64 output samples */

};
#endif

/* CodecMidi */
struct CodecMidi {
  /** Basic Codec structure. */
  //  struct Codec c;
  /** 64 mono samples long temporary audio data buffers */
    FLOAT cb1[OSCBUFFER];       /* cb1 no longer needs to be 1st element */
    FLOAT cb2[OSCBUFFER];
    FLOAT cb3[OSCBUFFER];       /* cb1..cb3 must still be together */

    Stereo buffer[OSCBUFFER];
    Stereo guitar30[OSCBUFFER]; /* buffer for overdriven guitar 30 */
    Stereo guitar31[OSCBUFFER]; /* buffer for distorted guitar 31 */
    s_int16 imaSamples[33];

    MidiChannel midich[CHANNELS];	/*MIDI channel information*/
    MidiNote notes[NUMNOTES];		/*Note information*/
    MIP MIPTable[CHANNELS];		/*Channel masking table*/

    s_int16 spMIDI;		/*"This is SP-MIDI file" flag*/
    s_int16 masterTranspose;
    s_int16 status;		/* status value (0x90 note-on etc.) */
    s_int16 polyphony;	/*Simultaneous notes (True value)*/
    s_int32 currentSample;	/*current time in samples*/
    u_int16 decodeTime;
    s_int32 tempo; /*Tempo microseconds/beat (default tempo is 120 beats/minute)*/

    FLOAT masterVolume;

    u_int16 division;	/*from the MThd (File header)*/
    s_int32 deltaToSamples;	/*Factor to convert delta time to samples*/

    s_int32 bytesLeft;		/*number of bytes to read*/
    s_int32 bytesTotal;
    s_int32 decodeSeconds;
  s_int32 totalSeconds;
    u_int16 bytesPerSecond;
    s_int16 playSpeed;

    s_int16 revType;      /*reverb type (size of the space: 0,1,2,3,4,8)*/
    s_int16 revTime;      /*reverb time (44(Type 0),501,2,3,4,8)*/
    FLOAT *revPhase;    /*&revBuffer[0] or &revBuffer[OSCBUFFER]*/
    FLOAT *revOutPhase; /*&revOut[0] or &revOut[REV_BUFFER]*/
    CombFilter combVal[4]; /*structure for comb filter intermediates*/
    FLOAT revPrev;
    FLOAT revTotalOut;
    FLOAT revBuffer[2*OSCBUFFER]; /* mono reverb buffer */
    FLOAT revOut[2*REV_BUFFER];   /* reverb intermediate output buffer */
    s_int16 revCurrent;   /* current decay time setting */
    s_int16 revUser;      /* the last user setting */
#if 0
  s_int16 chorBaseDelay; /* total delay = chorBaseDelay + sin(modPhase)*modDepth/2 */
#else
    s_int16 decodeStarted;
#endif
#ifdef CHORUS
  FLOAT chorModPhase; /* LFO phase of chorus */
  s_int16 chorWriteIdx; /* Index of chorus delay line write point */
  s_int16 chorType;  /*0:Chorus 1, 1:Chorus 2, 2:Chorus 3, 
					   3:Chorus 4, 5:FB Chorus, 5:Flanger*/
  FLOAT chorModRate; /* modulation frequency (value * 0.122 Hz)*/
  int chorModDepth; /* moudlation depth (peak-to-peak) (value + 1)/3.2 ms*/
  FLOAT chorFeedback; /* amount of feedback from chorus output (value * 0.763 %)*/
  FLOAT chorSendToReverb; /* send level from chorus to reverb in percent (0.787 *value)*/
  FLOAT chorBuffer[2*OSCBUFFER];
#endif/*CHORUS*/
    struct SampleInterp interpData[MAX_INTERP_SAMPLES];
    FLOAT compandVolume;
    FLOAT globalPhase[GLOBAL_PHASES];
  FLOAT distortDelay[16];
  s_int16 noDecode;
//  s_int32 ff;
  s_int32 dTime;
  s_int16 used;
  u_int32 fileLeftOld;
    u_int16 excludedChannels;
};

extern struct CodecMidi codecMidi;
extern FLOAT CodMidiGiveHPGain(MidiNote *note, register const int *p);


/* RealTime Midi support */
/* Prerequisites:
   1) Samplerate is 44100U
   2) UART speed is 31250bps
   3) RtMidiRxFill() RtMidiRxGet() must be provided
 */
void RtMidiMain(void); /* Main loop, does NOT check for outofwav! */
extern u_int16 RtMidiRxFill(void); /* Must be provided, returns # of bytes */
extern u_int16 RtMidiRxGet(void);  /* Must be provided, returns next byte */



#endif /* !ASM */

#endif

