#define STANDALONE
#include <module.h>
#include <stdlib.h>

#define PERIP(a) USEX(a)
void putword(register __a0 int ch);
void putch(register __a0 int ch);

/* Moved the invert check to pcmodflash.c so that the programmer
   does not need to know about it. It is then easier to change by just
   recompiling one program than to compile the whole chain and create
   autoprogimg.h. */

u_int16 errors = 0;

/************************************************************************/
#define SPI_EEPROM_COMMAND_WRITE_ENABLE  0x06
#define SPI_EEPROM_COMMAND_WRITE_DISABLE  0x04
#define SPI_EEPROM_COMMAND_READ_STATUS_REGISTER  0x05
#define SPI_EEPROM_COMMAND_WRITE_STATUS_REGISTER  0x01
#define SPI_EEPROM_COMMAND_READ  0x03
#define SPI_EEPROM_COMMAND_WRITE 0x02
#define SPI_EEPROM_COMMAND_CLEAR_ERROR_FLAGS 0x30
#define SPI_EEPROM_COMMAND_ERASE_SECTOR 0xD8
#define SPI_EEPROM_COMMAND_ERASE_4K_SECTOR 0x20
#define SPI_EEPROM_COMMAND_CHIP_ERASE 0x60

#define SPI_WREN  0x06
#define SPI_WRDI  0x04
#define SPI_RDSR  0x05
#define SPI_WRSR  0x01
#define SPI_READ  0x03
#define SPI_WRITE 0x02

#define SPI_xCS   1 /* GPIO */
#define SPI_CLK   2
#define SPI_MISO  4
#define SPI_MOSI  1 /* DREQ */

auto void SpiPrivDelay(void) {
    int i;
    for (i=0;i<10;i++) {
	USEX(SER_DREQ);
    }
}
auto u_int16 SpiPrivSendReceive(register __a0 u_int16 dat,
			    register __a1 s_int16 bits) {
    register u_int16 odata = 0;
    while (bits--) {
	USEX(SER_DREQ) = (dat>>bits);
	SpiPrivDelay();
	USEX(GPIO_ODATA) |= SPI_CLK;
	SpiPrivDelay();
	odata <<= 1;
#if SPI_MISO == 4
	odata |= (USEX(GPIO_IDATA) >> 2) & 1;
#else
	odata |= (USEX(GPIO_IDATA) & SPI_MISO) ? 1 : 0;
#endif
	USEX(GPIO_ODATA) &= ~SPI_CLK;
    }
    return odata;
}


void SpiPrivInit(void) {
    USEX(GPIO_ODATA) = SPI_xCS;
    USEX(GPIO_DDR) = SPI_xCS|SPI_CLK;
    USEX(GPIO_ODATA) = SPI_xCS;
}

u_int16 SpiPrivRead(u_int16 addr) {
    u_int16 dat;

    USEX(GPIO_ODATA) |= SPI_xCS;
    SpiPrivDelay();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(SPI_READ, 8);
    SpiPrivDelay();
    SpiPrivSendReceive(0, 8); //24-bit address
    SpiPrivSendReceive(addr, 16);
    SpiPrivDelay();
    dat = SpiPrivSendReceive(0, 16);
    SpiPrivDelay();
    USEX(GPIO_ODATA) |= SPI_xCS;
    return dat;
}

u_int16 SpiPrivStatus(void) {
    u_int16 dat;

    USEX(GPIO_ODATA) |= SPI_xCS;
    SpiPrivDelay();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(SPI_RDSR, 8);
    dat = SpiPrivSendReceive(0, 8);
    SpiPrivDelay();
    USEX(GPIO_ODATA) |= SPI_xCS;
    return dat;
}

u_int16 SpiPrivWrite(u_int16 addr, u_int16 wrdat) {
    u_int16 dat;

    USEX(GPIO_ODATA) |= SPI_xCS;
    SpiPrivDelay();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(SPI_WREN, 8);
    SpiPrivDelay();
    USEX(GPIO_ODATA) |= SPI_xCS;

    SpiPrivDelay();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(SPI_WRITE, 8);
    SpiPrivDelay();
    SpiPrivSendReceive(0, 8); //24-bit address
    SpiPrivSendReceive(addr, 16);
    SpiPrivDelay();
    dat = SpiPrivSendReceive(wrdat, 16);
    SpiPrivDelay();
    USEX(GPIO_ODATA) |= SPI_xCS;
    return dat;
}

/// Wait for not_busy (status[0] = 0) and return status
u_int16 SpiWaitStatus(void) {
  u_int16 status;
  USEX(GPIO_ODATA) |= SPI_xCS;
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_READ_STATUS_REGISTER,8);
  while ((status = SpiPrivSendReceive(0,8)) & 0x01){
    // puthex(status);puts("=status");
  }
  //Wait until ready
  USEX(GPIO_ODATA) |= SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;
  return status;
}


void SingleCycleCommand(u_int16 cmd){
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(cmd, 8);
    USEX(GPIO_ODATA) |= SPI_xCS;
}


#if 0
u_int16 SpiReadBlock(u_int16 blockn, u_int16 *dptr) {
  SpiWaitStatus();
  
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_READ,8);
  SpiPrivSendReceive(blockn>>7,8);            // Address[23:16] = blockn[14:7]
  SpiPrivSendReceive((blockn<<1)&0xff,8);     // Address[15:8]  = blockn[6:0]0
  SpiPrivSendReceive(0x00,8);                    // Address[7:0]   = 00000000
  //SPI_MASTER_16BIT_CSLO;
  {
    int n;
    for (n=0; n<256; n++){
      *dptr++ = SpiPrivSendReceive(0,16);
    }
  }
  USEX(GPIO_ODATA) |= SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;

  return 0;
}
#endif

u_int16 SpiVerifyBlock(u_int16 blockn, u_int16 *dptr) {
  SpiWaitStatus();
  
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_READ,8);
  SpiPrivSendReceive(blockn>>7,8);       // Address[23:16] = blockn[14:7]
  SpiPrivSendReceive((blockn<<1),8);     // Address[15:8]  = blockn[6:0]0
  SpiPrivSendReceive(0x00,8);            // Address[7:0]   = 00000000
  //SPI_MASTER_16BIT_CSLO;
  {
    int n;
    for (n=0; n<256; n++){
	u_int16 t = SpiPrivSendReceive(0,16);
	if (*dptr++ != t) {
#if 0
	    puthex(n);
	    puthex(dptr[-1]);
	    puthex(t);
	    puts("=!");
#endif
	    //SPI_MASTER_8BIT_CSHI;	
	    USEX(GPIO_ODATA) |= SPI_xCS;
	    return 1;
	}
    }
  }
  USEX(GPIO_ODATA) |= SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;
  return 0;
}


/// Write a block to EEPROM. Caution: Does not erase block
/// \param blockn number of sector to write: 0..32767 (16Mbytes)
/// \param dptr pointer to data block
u_int16 SpiWriteBlock(u_int16 blockn, u_int16 *dptr) {
  
  SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);
  SingleCycleCommand(SPI_EEPROM_COMMAND_CLEAR_ERROR_FLAGS);
  SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);

  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_WRITE_STATUS_REGISTER, 8);
  SpiPrivSendReceive(0x02, 8); //Sector Protections Off
  //SPI_MASTER_8BIT_CSHI;
  USEX(GPIO_ODATA) |= SPI_xCS;

  SpiWaitStatus();
  SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_WRITE, 8);
  SpiPrivSendReceive((blockn>>7),8);            // Address[23:16] = blockn[14:7]
  SpiPrivSendReceive((blockn<<1)&0xff,8);     // Address[15:8]  = blockn[6:0]0
  SpiPrivSendReceive(0,8);                    // Address[7:0]   = 00000000
 
  //SPI_MASTER_16BIT_CSLO;
  {
    u_int16 n;
    for (n=0; n<128; n++){
      SpiPrivSendReceive(*dptr++,16);
    }
  }
  //SPI_MASTER_8BIT_CSHI;
  USEX(GPIO_ODATA) |= SPI_xCS;

  SpiWaitStatus();

  SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_WRITE,8);
  SpiPrivSendReceive(blockn>>7,8);       
  SpiPrivSendReceive(((blockn<<1)+1)&0xff,8); // Address[15:8]  = blockn[6:0]1
  SpiPrivSendReceive(0,8);              
  
  //SPI_MASTER_16BIT_CSLO;
  {
    int n;
    for (n=128; n<256; n++){
      SpiPrivSendReceive(*dptr++,16);
    }
  }
  USEX(GPIO_ODATA) |= SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;
  SpiWaitStatus();
  return 0;
}


/*********************************************************************/


#define BUF_SIZE (256+3) //512 bytes
struct ENT {
    unsigned short cnt;
    unsigned short mode;
    unsigned short size;
    unsigned short block; //block address
    unsigned short data[BUF_SIZE];
};
__near struct ENT buffer[2];
__near struct ENT *rBuf, *wBuf;


//void SerReceive(void) {
auto void Interrupt0(void) {
    register __near short *ptr;
    register unsigned short cnt = rBuf->cnt;
    if (cnt == -1 || (cnt == 0 && PERIP(UART_DATA) != 0xbb)) {
	return; /* not the right start marker */
    }
    ptr = &rBuf->data[cnt/2];
    if (cnt & 1) {
	*ptr |= PERIP(UART_DATA);
    } else {
	*ptr = PERIP(UART_DATAH);
    }
    cnt++;
    if (cnt >= 6) {
	if (cnt == 6) {
	    /* get length and address */
	    rBuf->size    = rBuf->data[0] & 255;
	    rBuf->block   = rBuf->data[1];
	}
	if (cnt == 6 + 512) {
	    cnt = -1;
	}
    }
    rBuf->cnt = cnt;
}


int CheckErase(int szBits) {
  if (szBits) {
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    //SPI_MASTER_8BIT_CSLO;
    SpiPrivSendReceive(SPI_EEPROM_COMMAND_READ,8);
    SpiPrivSendReceive(0,8);            // Address[23:16] = blockn[14:7]
    SpiPrivSendReceive(0,8);     // Address[15:8]  = blockn[6:0]0
    SpiPrivSendReceive(0,8);                    // Address[7:0]   = 00000000
    {
      s_int32 n, sz = (1L<<szBits)>>1; /* in words */
      sz = 65536; //5120; //Check SD firmware area only
      for (n=0; n<sz; n++){
	if (SpiPrivSendReceive(0,16) != 0xffffU) {
	  USEX(GPIO_ODATA) |= SPI_xCS;
	  return 1;
	}
      }
    }
    USEX(GPIO_ODATA) |= SPI_xCS;
    return 0;
  }
  return 1; /* Chip Erase needed. */
}


/**
  Four interrupt stubs that the programmer can use without using ASM.
  For example to set up the first interrupt stub to GPIO0 interrupt,
  use the following:
  \code
    WriteIRam(0x20+INTV_GPIO0, ReadIRam((u_int16)InterruptStub0));
  \endcode

  Then you can enable GPIO0 interrupt and it will call your routine
  Interrupt0() whenever GPIO0 interrupt request is generated.
  \code
    PERIP(GPIO0_INT_FALL) |= DISPLAY_XCS;
    PERIP(INT_ENABLEL) |= INTF_GPIO0;
  \endcode
 */
void InterruptStub0(void);
void InterruptStub1(void); /**< Stub function that can be plugged to any interrupt vector. Calls Interrupt1(), which must be provided by the user. */
void InterruptStub2(void); /**< Stub function that can be plugged to any interrupt vector. Calls Interrupt2(), which must be provided by the user. */
void InterruptStub3(void); /**< Stub function that can be plugged to any interrupt vector. Calls Interrupt3(), which must be provided by the user. */
auto void Interrupt0(void); /**< Called by InterruptStub0. */
auto void Interrupt1(void); /**< Called by InterruptStub1. */
auto void Interrupt2(void); /**< Called by InterruptStub2. */
auto void Interrupt3(void); /**< Called by InterruptStub3. */

void WriteIRam(register __i0 addr, register __a u_int32 val);
u_int32 ReadIRam(register __i0 addr);

void main(void) {
  u_int16 sz = 0;
#if 1
    uartByteSpeed = 115200/10 * USEX(0xc00c); /*serMul*/
#endif
//    USEX(INT_ENABLE) = 0;
    SpiPrivInit(); /* Set SPI clock etc */
    USEX(SCI_CLOCKF) = 0xa000;
    SetClockFromSpi();
    USEX(UART_DIV) = UartDiv(intClock2KHz, uartByteSpeed);

    {
	register long z;
	for (z=0;z<1000000;z++) {
	    USEX(UART_STATUS);
	}
    }

#if 0
    /* higher UART speed -- not with PC! */
    clockX = 7; /*20111114: 8->7 = 3.5x*/
    SetRate(32000);
//    PERIP(UART_DIV) = 0x0120; //TODO: hand-written interrupt routine!
    //Note: currently UART is not the bottle-neck.
#endif
//    errors = 0;
    buffer[0].cnt = 0;
    buffer[1].cnt = 0;
    rBuf = &buffer[0];
    wBuf = &buffer[1];

    /*    USEY(WATCHDOG) = 0x2f;*/

    //PERIP(INT_ORIGIN) = INTF_RX; /* ack */
    //PERIP(UART_DATA); /* read dangling char? */
    //AddIntServer(9, (unsigned long)SerReceive);
    WriteIRam(0x20+INT_EN_RX, ReadIRam((u_int16)InterruptStub0));
    PERIP(INT_ENABLE) = (1<<INT_EN_RX);
    PERIP(INT_GLOB_ENA) = 0;

#if 0
    {
	u_int16 dat, da2;
	USEX(GPIO_ODATA) |= SPI_xCS;
	SpiPrivDelay();
	USEX(GPIO_ODATA) &= ~SPI_xCS;
	SpiPrivSendReceive(SPI_READ, 8);
	SpiPrivSendReceive(0, 16);
	dat = SpiPrivSendReceive(0, 16);
	da2 = SpiPrivSendReceive(0, 16);
	SpiPrivDelay();
	USEX(GPIO_ODATA) |= SPI_xCS;

	putword((''<<8)|(dat>>8));
	putword((''<<8)|(dat&0xff));
	putword((''<<8)|(da2>>8));
	putword((''<<8)|(da2&0xff));
	if (da2 != 0xffff && (dat & 0xff00)) {
	    goto failed; //Will have boot problems due to the extra bits..
	}
    }
#endif

    /*Check SPI memory*/
    {
	register u_int16 status;
	USEX(GPIO_ODATA) |= SPI_xCS;
	for (status=0;status<150;status++) {
	    //BusyWait10();
	    SpiPrivSendReceive(0,16);
	    SpiPrivSendReceive(0,16);
	    SpiPrivSendReceive(0,16);
	}

	USEX(GPIO_ODATA) &= ~SPI_xCS;
	SpiPrivSendReceive(0x9f,8);
	SpiPrivSendReceive(0x00,8);
	status = SpiPrivSendReceive(0x00,16);
	USEX(GPIO_ODATA) |= SPI_xCS;
	sz = (status & 255);

	if (1) { //status != 0 && status != 0xffffU) {
	    putword((''<<8)|(status & 0xff));
	    putword((''<<8)|0x40); //Ready
	} else {
	    goto failed;
	}
    }

#define REAL

#ifdef REAL
    //putstr("Erasing");

    /*may/must turn off PLL during chip erase!*/
    //BusyWait10();
#if 0
    while (PERIP(UART_STATUS) & (UART_ST_TXRUNNING|UART_ST_TXFULL))
	;
    PERIP(FREQCTLH) &= ~FCH_FORCE_PLL;
#endif

    if (CheckErase(sz)) {

    /* Chip erase -- Both in XCS and in CS2 */
      SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);
      SingleCycleCommand(SPI_EEPROM_COMMAND_CHIP_ERASE);
      SpiWaitStatus();
      //putstr("Done");
      putword((''<<8)|0x41); /* chip erase(s) complete */
    } else {
      putword((''<<8)|0x46); /* chip erase(s) skipped */
    }
    /*restore PLL after chip erase!*/
#if 0
    if (clockX > 2) {
	PERIP(FREQCTLH) |= FCH_FORCE_PLL;
    }
#endif
#else
    putword((''<<8)|0x41); /* chip erase(s) complete */
#endif

    putword((''<<8)|0x42); /* request first block */
    while (1) {
	if (rBuf->cnt == -1) {
	    struct ENT *tBuf = rBuf;
	    unsigned short chksum = 0xffff;
	    register unsigned short *ptr;
	    int i;

	    ptr = &tBuf->data[3];
	    for (i=0; i<256/*tBuf->length*/; i++) {
		chksum += *ptr++;
	    }
	    if (chksum != ~tBuf->data[2]) {
		rBuf->cnt = 0;   /* nak -- discard buffer */
		putword((''<<8)|0x43);
	    } else {
		wBuf->cnt = 0;
		rBuf = wBuf;

		wBuf = tBuf;
#ifdef REAL
		if (wBuf->size == 1) {
		    u_int16 *dat = &wBuf->data[3];
		    putword((''<<8)|0x42);/*ack -- program buffer, receive next*/
		    SpiWriteBlock(wBuf->block, dat);
		    if (SpiVerifyBlock(wBuf->block, dat)) {
			errors++;
			break;
		    }

		} else {
		    if (wBuf->size) {
			break; //0=ok, 255=EOF
		    }
		    putword((''<<8)|0x42);/*ack -- receive next*/
		}
#else
		putword((''<<8)|0x42);/*ack -- program buffer, receive next*/
#endif
	    }
	}
    }
    //puthex(errors>>16);
    //puthex(errors & 0xffff);
    //putstr(" errors\n");

//USEX(GPIO_ODATA) = 0; /*Drive low to test the pull-up resistor at reboot.*/
    if (!errors) {
	putword((''<<8)|0x45); /* finished ok */
	putword((''<<8)|0x45); /* finished ok */
    } else {
failed:
	putword((''<<8)|0x44); /* programming failed */
	putword((''<<8)|0x44); /* programming failed */
    }
    USEX(WDOG_CONFIG) = 10;
    USEX(WDOG_RESET) = 0x4ea9; /*Activate watchdog*/
    while (1)
	;
}

