/*

FILE: prom.c - SPI boot programmer
DESCRIPTION: 

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define STANDALONE
#include "vs1103.h"

#undef puts
#undef putchar

void putch(register __a0 short ch);

void Halt();
__near void vs3emubreak(register __i1 bp);


typedef long s_int32;
typedef short s_int16;
typedef unsigned short u_int16;
typedef unsigned long u_int32;
typedef __fract short f_int16;
typedef __fract long f_int32;

int largeAddr = 1;

static char hex[] = "0123456789abcdef";
void puthex(unsigned int a) {
    char tmp[5];
    tmp[0] = hex[(a>>12)&15];
    tmp[1] = hex[(a>>8)&15];
    tmp[2] = hex[(a>>4)&15];
    tmp[3] = hex[(a>>0)&15];
    tmp[4] = '\0';
    fputs(tmp, stdout);
}

//#define USEY(x) *((__y volatile unsigned short *)(unsigned short)(x))
#ifndef GPIO_IDATA
#define GPIO_IDATA GPIO_DATA
#endif

#define SPI_xCS   1 /* GPIO */
#define SPI_CLK   2
#define SPI_MISO  4
//#define SPI_xCS2  8
#define SPI_MOSI  1 /* DREQ */

#define SPI_WREN  0x06
#define SPI_WRDI  0x04
#define SPI_RDSR  0x05
#define SPI_WRSR  0x01
#define SPI_READ  0x03
#define SPI_WRITE 0x02

#define SPI_EEPROM_COMMAND_WRITE_ENABLE  0x06
#define SPI_EEPROM_COMMAND_WRITE_DISABLE  0x04
#define SPI_EEPROM_COMMAND_READ_STATUS_REGISTER  0x05
#define SPI_EEPROM_COMMAND_WRITE_STATUS_REGISTER  0x01
#define SPI_EEPROM_COMMAND_READ  0x03
#define SPI_EEPROM_COMMAND_WRITE 0x02
#define SPI_EEPROM_COMMAND_CLEAR_ERROR_FLAGS 0x30
#define SPI_EEPROM_COMMAND_ERASE_SECTOR 0xD8
#define SPI_EEPROM_COMMAND_ERASE_4K_SECTOR 0x20



auto void SpiPrivDelay(void) {
    int i;
    for (i=0;i<10;i++) {
	USEX(SER_DREQ);
    }
}
auto u_int16 SpiPrivSendReceive(register __a0 u_int16 dat,
			    register __a1 s_int16 bits) {
    register u_int16 odata = 0;
    while (bits--) {
	USEX(SER_DREQ) = (dat>>bits);
	SpiPrivDelay();
	USEX(GPIO_ODATA) |= SPI_CLK;
	SpiPrivDelay();
	odata <<= 1;
#if SPI_MISO == 4
	odata |= (USEX(GPIO_IDATA) >> 2) & 1;
#else
	odata |= (USEX(GPIO_IDATA) & SPI_MISO) ? 1 : 0;
#endif
	USEX(GPIO_ODATA) &= ~SPI_CLK;
    }
    return odata;
}


void SpiPrivInit(void) {
    USEX(GPIO_ODATA) = SPI_xCS;
    USEX(GPIO_DDR) = SPI_xCS|SPI_CLK;
    USEX(GPIO_ODATA) = SPI_xCS;
}

u_int16 SpiPrivRead(u_int16 addr) {
    u_int16 dat;

    USEX(GPIO_ODATA) |= SPI_xCS;
    SpiPrivDelay();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(SPI_READ, 8);
    SpiPrivDelay();
    if (largeAddr) {
	SpiPrivSendReceive(0, 8); //24-bit address
    }
    SpiPrivSendReceive(addr, 16);
    SpiPrivDelay();
    dat = SpiPrivSendReceive(0, 16);
    SpiPrivDelay();
    USEX(GPIO_ODATA) |= SPI_xCS;
    return dat;
}

u_int16 SpiPrivStatus(void) {
    u_int16 dat;

    USEX(GPIO_ODATA) |= SPI_xCS;
    SpiPrivDelay();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(SPI_RDSR, 8);
    dat = SpiPrivSendReceive(0, 8);
    SpiPrivDelay();
    USEX(GPIO_ODATA) |= SPI_xCS;
    return dat;
}

u_int16 SpiPrivWrite(u_int16 addr, u_int16 wrdat) {
    u_int16 dat;

    USEX(GPIO_ODATA) |= SPI_xCS;
    SpiPrivDelay();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(SPI_WREN, 8);
    SpiPrivDelay();
    USEX(GPIO_ODATA) |= SPI_xCS;

    SpiPrivDelay();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(SPI_WRITE, 8);
    SpiPrivDelay();
    if (largeAddr) {
	SpiPrivSendReceive(0, 8); //24-bit address
    }
    SpiPrivSendReceive(addr, 16);
    SpiPrivDelay();
    dat = SpiPrivSendReceive(wrdat, 16);
    SpiPrivDelay();
    USEX(GPIO_ODATA) |= SPI_xCS;
    return dat;
}

/// Wait for not_busy (status[0] = 0) and return status
u_int16 SpiWaitStatus(void) {
  u_int16 status;
  USEX(GPIO_ODATA) |= SPI_xCS;
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_READ_STATUS_REGISTER,8);
  while ((status = SpiPrivSendReceive(0,8)) & 0x01){
    // puthex(status);puts("=status");
  }
  //Wait until ready
  USEX(GPIO_ODATA) |= SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;
  return status;
}

void SingleCycleCommand(u_int16 cmd){
    USEX(GPIO_ODATA) |= SPI_xCS;
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(cmd, 8);
    USEX(GPIO_ODATA) |= SPI_xCS;
}


auto void AllInts(void) {
}

u_int16 SpiStatus(void) {
    register u_int16 dat;

    USEX(GPIO_ODATA) |= SPI_xCS;
    SpiPrivDelay();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(SPI_RDSR, 8);
    dat = SpiPrivSendReceive(0, 8);
    SpiPrivDelay();
    USEX(GPIO_ODATA) |= SPI_xCS;
    return dat;
}


#if 0
#define EEPROM_PAGE_SIZE_BITS 5
#define EEPROM_PAGE_SIZE (1<<EEPROM_PAGE_SIZE_BITS)

u_int16 SpiBlockWrite(u_int16 addr, register u_int16 *buf, u_int16 len) {
    USEX(GPIO_ODATA) |= SPI_xCS;
    SpiPrivDelay();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(SPI_WREN, 8);
    SpiPrivDelay();
    USEX(GPIO_ODATA) |= SPI_xCS;


    SpiPrivDelay();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(SPI_WRITE, 8);
    if (largeAddr) {
	SpiPrivSendReceive(0, 8); //24-bit address
    }
    SpiPrivSendReceive(addr, 16);
    while (len--) {
	SpiPrivSendReceive(*buf, 16);
	addr += 2;
	buf++;
    }
    SpiPrivDelay();
    USEX(GPIO_ODATA) |= SPI_xCS;
    return addr;
}

u_int16 SpiVerify(u_int16 addr, register u_int16 *buf, u_int16 len) {
    u_int16 dat;

    USEX(GPIO_ODATA) |= SPI_xCS;
    SpiPrivDelay();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(SPI_READ, 8);
    SpiPrivDelay();
    if (largeAddr) {
	SpiPrivSendReceive(0, 8); //24-bit address
    }
    SpiPrivSendReceive(addr, 16);
    SpiPrivDelay();

    while (len--) {
	if (SpiPrivSendReceive(0, 16) != *buf++)
	    return -1;
    }
    SpiPrivDelay();
    USEX(GPIO_ODATA) |= SPI_xCS;
    return 0;
}

u_int16 BlockProgram(u_int16 addr, u_int16 *buf, s_int16 words) {
    register s_int16 i, len = words;
    for (i=0;i<len;) {
	register u_int16 sz = (((addr&~(EEPROM_PAGE_SIZE-1))+EEPROM_PAGE_SIZE)
			       - addr) / 2;
	if (sz > len-i)
	    sz = len-i;
	SpiBlockWrite(addr, buf+i, sz);
	i += sz;
	addr += 2*sz;
	while (SpiPrivStatus() & 1)
	    SpiPrivDelay();
    }
    return SpiVerify(addr - 2*len, buf, len);
}
#endif



u_int16 SpiReadBlock(u_int16 blockn, u_int16 *dptr) {
  SpiWaitStatus();
  
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_READ,8);
  SpiPrivSendReceive(blockn>>7,8);            // Address[23:16] = blockn[14:7]
  SpiPrivSendReceive((blockn<<1)&0xff,8);     // Address[15:8]  = blockn[6:0]0
  SpiPrivSendReceive(0x00,8);                    // Address[7:0]   = 00000000
  //SPI_MASTER_16BIT_CSLO;
  {
    int n;
    for (n=0; n<256; n++){
      *dptr++ = SpiPrivSendReceive(0,16);
    }
  }
  USEX(GPIO_ODATA) |= SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;

  return 0;
}

/// Write a block to EEPROM. Caution: Does not erase block
/// \param blockn number of sector to write: 0..32767 (16Mbytes)
/// \param dptr pointer to data block
u_int16 SpiWriteBlock(u_int16 blockn, u_int16 *dptr) {
  
  SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);
  SingleCycleCommand(SPI_EEPROM_COMMAND_CLEAR_ERROR_FLAGS);
  SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);

  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_WRITE_STATUS_REGISTER, 8);
  SpiPrivSendReceive(0x02, 8); //Sector Protections Off
  //SPI_MASTER_8BIT_CSHI;
  USEX(GPIO_ODATA) |= SPI_xCS;

  SpiWaitStatus();
  SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_WRITE, 8);
  SpiPrivSendReceive((blockn>>7),8);            // Address[23:16] = blockn[14:7]
  SpiPrivSendReceive((blockn<<1)&0xff,8);     // Address[15:8]  = blockn[6:0]0
  SpiPrivSendReceive(0,8);                    // Address[7:0]   = 00000000
 
  //SPI_MASTER_16BIT_CSLO;
  {
    u_int16 n;
    for (n=0; n<128; n++){
      SpiPrivSendReceive(*dptr++,16);
    }
  }
  //SPI_MASTER_8BIT_CSHI;
  USEX(GPIO_ODATA) |= SPI_xCS;

  SpiWaitStatus();

  SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_WRITE,8);
  SpiPrivSendReceive(blockn>>7,8);       
  SpiPrivSendReceive(((blockn<<1)+1)&0xff,8); // Address[15:8]  = blockn[6:0]1
  SpiPrivSendReceive(0,8);              
  
  //SPI_MASTER_16BIT_CSLO;
  {
    int n;
    for (n=128; n<256; n++){
      SpiPrivSendReceive(*dptr++,16);
    }
  }
  USEX(GPIO_ODATA) |= SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;
  SpiWaitStatus();
  return 0;
}

u_int16 SpiVerifyBlock(u_int16 blockn, u_int16 *dptr) {
  SpiWaitStatus();
  
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_READ,8);
  SpiPrivSendReceive(blockn>>7,8);            // Address[23:16] = blockn[14:7]
  SpiPrivSendReceive((blockn<<1)&0xff,8);     // Address[15:8]  = blockn[6:0]0
  SpiPrivSendReceive(0x00,8);                    // Address[7:0]   = 00000000
  //SPI_MASTER_16BIT_CSLO;
  {
    int n;
    for (n=0; n<256; n++){
	u_int16 t = SpiPrivSendReceive(0,16);
	if (*dptr++ != t) {
	    puthex(n);
	    puthex(dptr[-1]);
	    puthex(t);
	    puts("=!");
	//SPI_MASTER_8BIT_CSHI;	
	USEX(GPIO_ODATA) |= SPI_xCS;
	return 1;
      }
    }
  }
  USEX(GPIO_ODATA) |= SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;
  return 0;
}


#if 1
void DumpBlock (u_int16 *dptr) {
  int n;
  for (n=0; n<256; n++){
    puthex(*dptr++);
    if ((n&7)==7) {
	puts("");
    }
    //put2c(*dptr++);
  }
  puts("");
}
#endif


/// Write a block to EEPROM. Caution: Does not erase block
/// \param blockn number of sector to write: 0..32767 (16Mbytes)
/// \param dptr pointer to data block

#if 0
/// Erase one erasable block
/// Erase size for intel flash: 64K (128 disk sectors)
void SpiEraseBlock(u_int16 blockn_inside_sector){
  SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);  
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_WRITE_STATUS_REGISTER,8);
  SpiPrivSendReceive(0x02,8); //Sector Protections Off
  USEX(GPIO_ODATA) |= SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;
  SpiWaitStatus();
  SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);  
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_ERASE_SECTOR,8);
  SpiPrivSendReceive(blockn_inside_sector>>7,8);       
  SpiPrivSendReceive((blockn_inside_sector<<1)&0xff,8);
  SpiPrivSendReceive(0,8);              
  USEX(GPIO_ODATA) |= SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;
  SpiWaitStatus();
}
#endif

void EeUnprotect(){
  SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  //SPI_MASTER_8BIT_CSLO;
  SpiPrivSendReceive(SPI_EEPROM_COMMAND_WRITE_STATUS_REGISTER,8);
  SpiPrivSendReceive(0x02,8); //Sector Protections Off
  USEX(GPIO_ODATA) |= SPI_xCS;
  //SPI_MASTER_8BIT_CSHI;
  SpiWaitStatus();
  SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);
}

void Erase4kBlock(u_int16 blockn) {
    // Erase 4K sector
    SingleCycleCommand(SPI_EEPROM_COMMAND_WRITE_ENABLE);
    SingleCycleCommand(SPI_EEPROM_COMMAND_CLEAR_ERROR_FLAGS);
    EeUnprotect();
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    //SPI_MASTER_8BIT_CSLO;
    SpiPrivSendReceive(SPI_EEPROM_COMMAND_ERASE_4K_SECTOR,8);
    SpiPrivSendReceive(blockn>>7,8);            // Address[23:16] = blockn[14:7]
    SpiPrivSendReceive((blockn<<1)&0xff,8);     // Address[15:8]  = blockn[6:0]0
    SpiPrivSendReceive(0,8);                    // Address[7:0]   = 00000000
    USEX(GPIO_ODATA) |= SPI_xCS;
    //SPI_MASTER_8BIT_CSHI;
    SpiWaitStatus();
}


void Rems() {
    int rdid[2];
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiPrivSendReceive(0x90, 8);
    SpiPrivSendReceive(0,16);
    SpiPrivSendReceive(0,8);
    rdid[0] = SpiPrivSendReceive(0,16);
    USEX(GPIO_ODATA) |= SPI_xCS;
    puthex(rdid[0]);
    puts(" =REMS"); /*c814*/
}




/*

FUNCTION: main()
DESCRIPTION: RTOS, peripheral and variable initialization.
INPUTS: none
RESULT: none
SEE ALSO: 

*/
#include "spifs.h"

#define BUF_SIZE (512/2)
static u_int16 buf[BUF_SIZE];
void MyMain(void) {
    FILE *fp;
    int i, len, blk = 0;

    USEX(INT_ENABLE) = 0;
    SpiPrivInit(); /* Set SPI clock etc */
    puthex(SpiPrivStatus()); /* should be 0080 */
    puts(" status");
    Rems();

//#define VERIFY_ONLY
//    SpiReadBlock(0, buf);
//    DumpBlock(buf);

    fp = fopen("spifs.img", "rb");
    if (!fp) {
	puts("Could not open spifs.img!");
	goto exit;
    }
    blk = SPIFSSTARTBLOCK;
    while (len = fread(buf, 1, BUF_SIZE, fp)) {
	if ((blk & 7)==0) {
	    puthex(blk);
#if 0 /* only header */
	    if(blk==SPIFSSTARTBLOCK+8)
		break;
#endif
#ifndef VERIFY_ONLY
	    puts(" Erase");
	    Erase4kBlock(blk);
#else
	    puts(" Chk");
#endif
	}
#ifndef VERIFY_ONLY
	SpiWriteBlock(blk, buf);
#endif
	if (SpiVerifyBlock(blk, buf)) {
	    puts(" Error");
	}
	memset(buf, -1, sizeof(buf));
	blk++;
#if 0
	if (blk >= 4096) { /* 2MB FLASH */
	    puts("Limited to 2MB!\n");
	    break;
	}
#endif
#if 0
	if (blk >= 8192) { /* 4MB FLASH */
	    puts("Limited to 4MB!\n");
	    break;
	}
#endif

    }
    fclose(fp);

//    puts("\nFinished!!\n");
//    fflush(stdout);

#if 0
    while (blk--) {
	SpiReadBlock(blk, buf);
	DumpBlock(buf);
    }
#endif

#if 1
 exit:
    {
	int i;
	for (i=0;i<61;i++)
	    putch(0x10);
	while ((USEX(UART_STATUS) & UART_ST_TXRUNNING))
	    ;
    }
#endif
    ((void (*)(void))(0x4000))();
}
