/*
  Note:

  DREQ is changed in normal firmware both in SDI interrupts when the
  stream buffer becomes too full, and whenever the stream buffer becomes
  empty enough.

  Because we have our own SDI handler, the interrupt does not touch DREQ.
  So, we can use DREQ ourselves as we wish and access to it does not need to
  be protected by interrupt Disable()/Enable(). However, the normal decoding
  will still keep raising DREQ when it sees that there is space for new data.

 */

#define STANDALONE
#define USE_EARSPEAKER
#include <string.h>
#include "module.h"
#include <parametric.h>
#include <general.h>

auto void PatchInitHardware(void);

void putch(register __a0 int ch);
void putch15(register __a0 int ch);
void putword(register __a0 int ch);
void Monitor(void); /*never returns*/
auto void play_frame(void);

//void Disable();
//void Enable();
//void Reset();

//#undef putchar
//#undef puts

//#include <stdio.h>

#include <stdlib.h>

#ifndef GPIO_IDATA
#define GPIO_IDATA GPIO_DATA
#endif

#define SPI_xCS   1 /* GPIO */
#define SPI_CLK   8 /*MMC clk*/
#define SPI_MISO  4
#define SPI_MISO_SHIFT -3
#define SPI_xCS2  2
#define SPI_MOSI  1 /* DREQ */

extern __y u_int16 hwSampleRate;


#if defined(USE_BUTTON)
__y struct {
    u_int16 Data; /* must be first entry! Used by c.s */
    u_int16 Cnt;
    u_int16 Strobe;
    u_int16 Timer; /* Data,Cnt,Strobe,Timer must be in this order (buts.s) */
    u_int16 Key;
    s_int16 KeyTime;
#ifdef KEYTIMER
    s_int16 KeyTimer;
#endif
} ui;
#endif

#if 1

#if defined(USE_BUTTON)
s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
		 register __a0 s_int16 n);
#endif/*USE_BUTTON*/

#else/*1*/

#if defined(USE_BUTTON)
s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
		 register __a0 s_int16 n) {
    /*Don't check mode, all we get is APPL_AUDIO.*/
    //if (mode == APPL_AUDIO)
    {
	register u_int16 t = hwSampleRate/16;
	/* generate a timebase for User Interface */
	ui.Cnt += n;
	if (ui.Cnt > t) {
	    ui.Cnt -= t;
	    ui.Timer++;
	    ui.Strobe = 1;
	}
    }
}
#endif/*USE_BUTTON*/
#endif

u_int16 ClockSdi(void); /* dummy return value */
#if 0
u_int16 ClockSdi(void) {
    register int i;
    for (i=0;i<32;i++) {
	USEX(SER_DREQ) = 1;
	USEX(SER_DREQ) = 0;
    }
    return -1;
}
#endif


/*
  Called before HALT
 */
void UserHook(void) {

#ifdef USE_BUTTON
    if (ui.Strobe) {
	ui.Strobe = 0;
	ui.Data = (USEX(GPIO_IDATA) & (1<<3));
	if (ui.Data == ui.Key) {
	    if (ui.KeyTime < 12) {
		ui.KeyTime++;
	    } else if (ui.Key && ui.KeyTime > 0) {
		/* long presses */
		if (ui.Key && ui.KeyTime == 12/*one-shot*/) {
		    ui.KeyTime = 13;
		    /* Volume toggle */
		    USEX(SCI_VOL) ^= 0x1818;
		    SetVolumeFromSpi();
		}
	    }
	} else {
	    if (ui.Data == 0 && ui.KeyTime > 0) {
		/* short presses */
		if (ui.KeyTime < 12) {
		key:
		    /* was pressed < 0.75 seconds */
		    if (ui.Key) {
#if 0 /* for pause-after-play */
			if (USEX(SCI_AICTRL3) & CTRL3_PAUSE_ON) {
			    /* pause -> next file */
			    USEX(SCI_AICTRL0) += 0x8001;
			} else {
			    /* play -> pause */
			    USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
			}
#else /* for pause-before-play */
			if (USEX(SCI_AICTRL3) & CTRL3_PAUSE_ON) {
			    /* pause -> play */
			    USEX(SCI_AICTRL3) &= ~CTRL3_PAUSE_ON;
			} else {
			    /* play -> next file */
			    USEX(SCI_AICTRL0) += 0x8001;
			}
#endif
		    }
		}
	    }
	    ui.Key = ui.Data;
	    ui.KeyTime = 0;
	}
    }
#endif
}


extern s_int16 rtmidi_int_vecs; /* Actually in code memory! */

auto void MyMain(register s_int16 __a1 fullReset) {
#ifdef BATTCHECK
    USEX(MODU_DIV) = 0x0030;
#endif/*BATTCHECK*/

#if defined(SCI_UI) && defined(ENABLE_WATCHDOG)
    USEX(WDOG_CONFIG) = 1000;
    USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif

#if defined(SCI_UI)
    USEX(SCI_AICTRL2) = 0;
#endif
    USEX(SCI_MODE) &= ~(1<<SCIMB_OUT_OF_WAV);
#if 0 //For vs1103 greeting card player
    USEX(SCI_MODE) |= (1<<SCIMB_DIFFERENTIAL);
#endif
#ifdef USE_BUTTON
    USEX(GPIO_DDR)   = SPI_xCS|SPI_xCS2;
    USEX(GPIO_ODATA) = SPI_xCS|SPI_xCS2;
#else
    USEX(GPIO_DDR)   = SPI_xCS|SPI_CLK|SPI_xCS2;
    USEX(GPIO_ODATA) = SPI_xCS|SPI_xCS2;
#endif

#if 1
    /* 'wake up' spi memory */
//    USEX(GPIO_ODATA) &= ~SPI_xCS;
//    SpiSendReceiveSpi((SPI_READ<<8), 8);
//    SpiSendReceiveSpi(sector<<1, 16);
    SpiSendReceiveSpi(0, 8);
//    USEX(GPIO_ODATA) |= SPI_xCS;
#endif


    PatchInitHardware(); /* does not clear user area */
    InitCommon();
    //AudioInit(); //rather use..
    //AudioSet(2, 44100U);

#if defined(USE_BUTTON)
    applAddr = ApplAddr;
#endif/*USE_BUTTON*/

    /* Allow DREQ and enable interrupts, VERY IMPORTANT! */
    USEX(INT_ENABLE) = 0;
//    USEX(INT_GLOB_ENA) = 0;
    USEX(INT_GLOB_ENA) = 0;
//    USEX(INT_GLOB_ENA) = 0;

    AudioSet(2, 44100U);
    InitReadFrame();

    USEX(INT_ENABLE) = INT_EN_ALL|(1<<INT_EN_MODU);
    USEX(SER_DREQ) = 1;
#if defined(ENABLE_WATCHDOG) && defined(KEEP_DRIVERS_OFF)
    /*otherwise keep it in powerdown until MMC found*/
    USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1); /* enable ANALOG power */
#endif

  /* Real Time Midi activated if GPIO0 = 0 and GPIO1 = 1 */
    {
	register u_int32 t;

	skipMixer = 1;
	USEX(SCI_MODE) = (USEX(SCI_MODE) & ~(SCIMF_EARSPEAKER_HILO))
	  #if 0
	  | ((1 << SCIMB_EARSPEAKER_LO) | (1 << SCIMB_EARSPEAKER_HI))
	  #endif
	  ;

	t = ReadFromProgramRam((u_int32)(&rtmidi_int_vecs));
	WriteToProgramRam(0x22, t>>16, t);
	t = ReadFromProgramRam((u_int32)(&rtmidi_int_vecs+1));
	WriteToProgramRam(0x25, t>>16, t);

	uartByteSpeed = 3125;
	SetClockFromSpi();

	RtMidiMain();
    }
}
