/*
  Note:

  DREQ is changed in normal firmware both in SDI interrupts when the
  stream buffer becomes too full, and whenever the stream buffer becomes
  empty enough.

  Because we have our own SDI handler, the interrupt does not touch DREQ.
  So, we can use DREQ ourselves as we wish and access to it does not need to
  be protected by interrupt Disable()/Enable(). However, the normal decoding
  will still keep raising DREQ when it sees that there is space for new data.

 */

#ifdef SCI_UI
//#define USE_BUTTON /* Wait for button press (in GPIO3) to play a file. */
//#define PAUSE_AT_POWERON /*switches pause-before-play to pause-after-play*/
#endif

#if 0
#define LOUDNESS_ALWAYS_ON_OFF
#endif


/*
  - subdirectory support

  NEW features for SCI_UI:
  - play mode:
    + normal continuous play
    + pause before song (doubles as pause after reset)
    + loop song
    + report file found / not found and wait for play command
  - find file by number but have the file name readable from memory

  TODO features:

  Extra features
  - L/R VU-meter through SCI
 */


#define STANDALONE
#define USE_EARSPEAKER
#include <string.h>
#include "module.h"
#include <parametric.h>
#include <general.h>
#ifdef UART_FIFO
#include "uartctrl.h"
#endif

//void SetSourcesFromSpi(void);
//auto void SetDreqSafetyAreaFromSpi(void);

void putch(register __a0 int ch);
void putch15(register __a0 int ch);
void putword(register __a0 int ch);
void Monitor(void); /*never returns*/

#if 0
static char hex[] = "0123456789ABCDEF";
void puthex(u_int16 d) {
    register i;
    for (i=0;i<4;i++) {
	putch(hex[((d>>12)&15)]);
	d <<= 4;
    }
}
#endif

void putstrp(register __i2 const char *s) {
    while (1) {
	if (!(*s & 0xff00))
	    break;
	putch15(*s);
	if (!(*s & 0xff))
	    break;
	putch(*s++);
    }
}

auto void play_frame(void);
//void Disable();
//void Enable();
//void Reset();

//#undef putchar
//#undef puts

//#include <stdio.h>

#include <stdlib.h>

#ifdef WAIT_FOR_UART_BYTE
#define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC))
#define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC))
#else
#define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#endif/*elseWAIT_FOR_UART_BYTE*/


#define KEEP_DRIVERS_OFF /* .. until MMC located */
#define WITH_SUBDIRECTORIES /* support subdirectories */

#define CTRL3_UPDATE_VOL        (1<<15)
#define CTRL3_I2S_ENABLE        (1<<9) /*VS1053: enable I2S*/
#define CTRL3_BY_NAME           (1<<8) /*locate by name*/
#define CTRL3_RECORD_ON         (1<<7)
#define CTRL3_AT_END            (1<<6)
#define CTRL3_NO_NUMFILES       (1<<5)
#define CTRL3_PAUSE_ON          (1<<4)
#define CTRL3_FILE_READY        (1<<3)

#define CTRL3_PLAY_MODE_MASK    (3<<1)
#define CTRL3_PAUSE_AFTER_PLAY  (3<<1)
#define CTRL3_PAUSE_BEFORE_PLAY (2<<1)
#define CTRL3_LOOP_SONG         (1<<1)
#define CTRL3_PLAY_NEXT         (0<<1)

#define CTRL3_RANDOM_PLAY       (1<<0) /* SCI_AICTRL3 */

#ifndef GPIO_IDATA
#define GPIO_IDATA GPIO_DATA
#endif

#define SPI_xCS   1 /* GPIO */
#define SPI_CLK   8 /*MMC clk*/
#define SPI_MISO  4
#define SPI_MISO_SHIFT -3
#define SPI_xCS2  2
#define SPI_MOSI  1 /* DREQ */

//#define SPI_WREN  0x06
//#define SPI_WRDI  0x04
//#define SPI_RDSR  0x05
//#define SPI_WRSR  0x01
#define SPI_READ  0x03
//#define SPI_WRITE 0x02

typedef enum {
    spi_pause=0,
    spi_file,
    spi_seek,
    spi_waitdata,
    spi_data,
    spi_zeros,
    spi_record,
    spi_waitima
} SPISTATE;

SPISTATE spiState, oldSpiState;
s_int16 spiCnt;

#include "spifs.h"
auto void ReadDiskSector(register __a u_int32 sector) {
    register s_int16 i;
    register u_int16 *buffer = stream_buffer;

    sector += SPIFSSTARTBLOCK;
#if 0
    puthex(sector>>16);
    puthex(sector);
    putch(' ');
    putch('s');
    putch('e');
    putch('c');
    putch('\n');
#endif

    USEX(GPIO_ODATA) |= SPI_xCS;
    USEX(GPIO_DDR)   |= SPI_xCS;
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiSendReceiveSpi((SPI_READ<<8), 8);
    SpiSendReceiveSpi(sector<<1, 16);
    SpiSendReceiveSpi(0, 8);
    for (i=0;i<512/2;i++) {
	*buffer++ = SpiSendReceiveSpi(0, 16);
#if 0
  puthex(buffer[-1]);
  if ((i & 7) == 7)
    putch('\n');
#endif
    }
    USEX(GPIO_ODATA) |= SPI_xCS;
    return;
}

auto u_int16 GetByte(register __c0 u_int16 n);
auto u_int16 GetWord(register __c0 u_int16 n);
auto u_int32 GetLong(register __c0 u_int16 n);

extern __y u_int16 fileName[6];

auto u_int16 InitFileSystem(void) {
//    ReadDiskSector(0);
    return 0;
}

u_int16 gFileNum[2];
u_int32 fileSize, fileStart;
u_int32 spiSize;
u_int32 spiAddr;

auto s_int16 OpenFile(register __c0 u_int16 fileNum) {
    register __c1 int i = 12;
    u_int32 currentSector;
    __y u_int32 dataStart;

    gFileNum[0] = fileNum;
    gFileNum[1] = 0;

#ifdef ENABLE_WATCHDOG
    USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
    ReadDiskSector(currentSector = 0);
    dataStart = GetLong(4);
    while (1) {
	register __d1 u_int16 fn = GetByte(i+0); /* first char of filename */

	/* We are now looking at FAT directory entry. */
	/* Is current file a regular file? */
	if (fn == 0 || fn == 0xff) {
	    goto eod; /*end of directory*/
	}
#if defined(SCI_UI)
	if (USEX(SCI_AICTRL3) & CTRL3_BY_NAME) {
	    register int j, t = 0;
	    register u_int16 *p = stream_buffer + i/2;
	    register __y u_int16 *d = fileName;
	    p[5] &= 0xff00; /* clear the attribute byte */
	    for (j=0; j<6; j++) {
		t |= *d++ ^ *p++;
	    }
	    if (t == 0) {
		USEX(SCI_AICTRL0) = gFileNum[1];
		goto thisfile;
	    }
	}
#endif

	gFileNum[1]++;
	if (gFileNum[0]-- == 0) {
	thisfile:
#ifdef SCI_UI
	    /* Put the 8.3-character filename to memory so the user
	       can read it. Two bytes per word, big-endian. */
	    {   /* 13 words */
		register int j;
		register u_int16 *p = stream_buffer + i/2;
		register __y u_int16 *d = fileName;
		for (j=0;j<6;j++) {
		    *d++ = *p++;
		}
	    }
#endif
	    /* ------------ FILE FOUND ------------- */
	    fileSize = GetLong(i+16);
	    fileStart = dataStart + GetLong(i+12) + SPIFSSTARTBLOCK;
#if 0
	    puthex(fileSize);
	    putch('s');
	    putch('z');
	    putch('\n');
#endif
	    return -1; /* File found, All OK return */
	}
	i = (i + 20);
	if (i >= 512) {
	    currentSector++;
	    i = 12;
	    if (currentSector >= dataStart) {
	eod:
#if 0
	    putch('e');
	    putch('o');
	    putch('d');
	    putch('\n');
#endif
#if defined(SCI_UI)
		USEX(SCI_AICTRL3) &= ~CTRL3_BY_NAME;
#endif
		return gFileNum[1];
	    }
	    ReadDiskSector(currentSector);
	}
    }
}


extern __y u_int16 hwSampleRate;
u_int16 __x * __x stream_save;

#if defined(USE_BUTTON)
__y struct {
    u_int16 Data; /* must be first entry! Used by c.s */
    u_int16 Cnt;
    u_int16 Strobe;
    u_int16 Timer; /* Data,Cnt,Strobe,Timer must be in this order (buts.s) */
    u_int16 Key;
    s_int16 KeyTime;
#ifdef KEYTIMER
    s_int16 KeyTimer;
#endif
} ui;
#endif

#if 1

#if defined(USE_BUTTON)
s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
		 register __a0 s_int16 n);
#endif/*USE_BUTTON*/

#else/*1*/

#if defined(USE_BUTTON)
s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
		 register __a0 s_int16 n) {
    /*Don't check mode, all we get is APPL_AUDIO.*/
    //if (mode == APPL_AUDIO)
    {
	register u_int16 t = hwSampleRate/16;
	/* generate a timebase for User Interface */
	ui.Cnt += n;
	if (ui.Cnt > t) {
	    ui.Cnt -= t;
	    ui.Timer++;
	    ui.Strobe = 1;
	}
    }
}
#endif/*USE_BUTTON*/
#endif

u_int16 ClockSdi(void); /* dummy return value */
#if 0
u_int16 ClockSdi(void) {
    register int i;
    for (i=0;i<32;i++) {
	USEX(SER_DREQ) = 1;
	USEX(SER_DREQ) = 0;
    }
    return -1;
}
#endif

s_int16 hcShift;
#ifdef UART_FIFO
u_int16 uiNum = 0;
s_int16 uiLoops = -1;
#endif/*UART_FIFO*/

/*
  Called before HALT
 */
void UserHook(void) {
    register __c1 SPISTATE cachedSpiState = spiState;

//USEX(UART_DATA) = spiState;

#if defined(UART_FIFO)
    if (UartFill()) {
	register int cmd = UartGetByte();
	if (cmd == 0xef) { /* Jump to monitor */
	    Disable();
	    Monitor();
	} else if (cmd == 'V') { /* Set direct volume */
	    USEX(SCI_VOL) = (uiNum & 255) * 0x0101;
	    uiNum = 0;
	    goto showvol;
	} else if (cmd == 'P') { /* play file num */
	    USEX(SCI_AICTRL0) = uiNum;
	    uiNum = 0;
	    /*Note: \n is not received, because we restart the software.
	            So, we clear uiNum here. */
	    goto cancelpauseoff;
	} else if (cmd == '\n' || cmd == '\r') {
	    uiNum = 0;
	} else if (cmd >= '0' && cmd <= '9') {
	    uiNum = uiNum*10 + cmd - '0';
	} else if (cmd == 'f') {
	    /* file mode */
	    USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
	    USEX(SCI_AICTRL3) |= CTRL3_PAUSE_BEFORE_PLAY;
//	    putch('f');
	} else if (cmd == 'c') {
	    /* continuous play mode */
	    USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
//	    putch('c');
	} else if (cmd == 'l') {
	    USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
	    USEX(SCI_AICTRL3) |= CTRL3_LOOP_SONG;
	    if (uiNum) {
		uiLoops = uiNum; //loop this many times
	    }
	    uiNum = 0;
//	    putch('l');
	} else if (cmd == 'C') { /* cancel play: replay the same file */
//	    putch('C');
	    goto cancelpauseoff;
	    //cs.cancel = 1;
	    //player.pauseOn = 0;
	} else if (cmd == '+') {
	    if (USEX(SCI_VOL)) {
		USEX(SCI_VOL) -= 0x0101; //KeyEventHandler(ke_volumeUp2);
	    }
	    goto showvol;//putword(player.volume);
	} else if (cmd == '-') {
	    //KeyEventHandler(ke_volumeDown2);
	    USEX(SCI_VOL) += 0x0101;
	showvol:
	    SetVolumeFromSpi();
	    putch(USEX(SCI_VOL));
	} else if (cmd == '=') {
	    /* pause */
#ifdef SCI_UI
	    USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
#else
	    /* play pause/continue -- not very good for midi.. */
	    if (cachedSpiState != spi_pause) {
		oldSpiState = cachedSpiState;
		cachedSpiState = spi_pause;
		stream_save = stream_wr_pointer;
		stream_wr_pointer = stream_rd_pointer+1;
	    }
#endif/*else SCI_UI*/
	    //player.pauseOn = 1;
	    goto echo;//putch('=');
	} else if (cmd == '>') {
	    /* play */
#ifdef SCI_UI
	    USEX(SCI_AICTRL3) &= ~CTRL3_PAUSE_ON;
#else
	    if (cachedSpiState == spi_pause) {
		cachedSpiState = oldSpiState;
		/* restore 'stolen' data */
		stream_wr_pointer = stream_save;
	    }
#endif
	    //player.pauseOn = 0;
	    parametric_x.midiConfig = 0; /* play speed 0 and reverb auto */
	    goto echo;//putch('>');
#if 1
	} else if (cmd == '') {
	    /* Fast forward */
	    //if (cs.fastForward < 4)
	    //++cs.fastForward;
	    /* increase play speed */
	    if ((parametric_x.midiConfig & 0x70) < 0x70)
		parametric_x.midiConfig += 0x10;
	    putch(((s_int16)parametric_x.midiConfig>>4) & 7);
#endif
	} else if (cmd == 'n') {
	    /* next */
	    //player.nextFile = player.currentFile + 1;
	    USEX(SCI_AICTRL0)++;
	    putch('n');
#if 1
	    goto cancelpauseoff;
	    //cs.cancel = 1;
	    //player.pauseOn = 0;
	} else if (cmd == 'p') {
	    /* previous */
	    //player.nextFile = player.currentFile - 1;
	    if (USEX(SCI_DECODE_TIME) < 5) {
		if (USEX(SCI_AICTRL0)) {
		    USEX(SCI_AICTRL0)--;
		} else {
		    USEX(SCI_AICTRL0) = USEX(SCI_AICTRL2) - 1;
		}
	    }
	    putch('p');
#endif
	cancelpauseoff:
	    //cs.cancel = 1;
	    //player.pauseOn = 0;
#ifdef SCI_UI
	    USEX(SCI_AICTRL0) |= 0x8000U;
#else
	    ;
#endif
	} else if (cmd == '?') {
	    /* position + other info -- Note: will block. */
	    putword(USEX(SCI_DECODE_TIME));
	    putch((fileSize-2*spiSize) / ((fileSize >> 8)+1));
	} else {
	    /*unknown command*/
	echo:
	    putch(cmd);
	}
    }
#endif/*UART_FIFO*/

#if defined(UART_UI)
    if ((USEX(SCI_AICTRL3) & CTRL3_UPDATE_VOL)) {
	USEX(SCI_AICTRL3) &= ~CTRL3_UPDATE_VOL;
	SetVolumeFromSpi();
    }
#endif/*UART_UI*/

#ifdef SCI_UI
    /* play pause/continue */
    if ((USEX(SCI_AICTRL3) & CTRL3_PAUSE_ON)) {
	if (cachedSpiState != spi_pause) {
	    oldSpiState = cachedSpiState;
#if defined(UART_UI) || defined(UART_FIFO)
	    putch('p');//pause
#endif
	    cachedSpiState = spi_pause;
	    stream_save = stream_wr_pointer;
	    stream_wr_pointer = stream_rd_pointer+1;
	    /* Does not matter if it is one off..
	       StreamDiff() returns the right value regardless. */
	}
    } else {
	if (cachedSpiState == spi_pause) {
#if defined(UART_UI) || defined(UART_FIFO)
	    putch('c');//continue
#endif
	    cachedSpiState = oldSpiState;
	    stream_wr_pointer = stream_save;
	}
    }
    if ((USEX(SCI_AICTRL0) & 0x8000U)) {
	/* new song */
	USEX(SCI_AICTRL0) &= 0x7fff; /* ack command */
	/*NEW: clear pause and file ready bits automatically */
	USEX(SCI_AICTRL3) &= ~(CTRL3_PAUSE_ON | CTRL3_FILE_READY | CTRL3_AT_END);
	USEX(GPIO_ODATA) |= SPI_xCS;
#ifdef UART_UI
	putch('C');//USEX(UART_DATA) = 'C';
#endif
	/* Jump to new track immediately */
#if defined(UART_UI) || defined(UART_FIFO)
	putstrp("\pdone\n\n");
#endif
	MyReset();
    }
#endif/*SCI_UI*/

#ifdef USE_BUTTON
    if (ui.Strobe) {
	ui.Strobe = 0;
	ui.Data = (USEX(GPIO_IDATA) & (1<<3));
	if (ui.Data == ui.Key) {
	    if (ui.KeyTime < 12) {
		ui.KeyTime++;
	    } else if (ui.Key && ui.KeyTime > 0) {
		/* long presses */
		if (ui.Key && ui.KeyTime == 12/*one-shot*/) {
		    ui.KeyTime = 13;
		    /* Volume toggle */
		    USEX(SCI_VOL) ^= 0x1818;
		    SetVolumeFromSpi();
		}
	    }
	} else {
	    if (ui.Data == 0 && ui.KeyTime > 0) {
		/* short presses */
		if (ui.KeyTime < 12) {
		key:
		    /* was pressed < 0.75 seconds */
		    if (ui.Key) {
#if 0 /* for pause-after-play */
			if (USEX(SCI_AICTRL3) & CTRL3_PAUSE_ON) {
			    /* pause -> next file */
			    USEX(SCI_AICTRL0) += 0x8001;
			} else {
			    /* play -> pause */
			    USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
			}
#else /* for pause-before-play */
			if (USEX(SCI_AICTRL3) & CTRL3_PAUSE_ON) {
			    /* pause -> play */
			    USEX(SCI_AICTRL3) &= ~CTRL3_PAUSE_ON;
			} else {
			    /* play -> next file */
			    USEX(SCI_AICTRL0) += 0x8001;
			}
#endif
		    }
		}
	    }
	    ui.Key = ui.Data;
	    ui.KeyTime = 0;
	}
    }
#endif

#define SW1 0xffffU
#define SW2 0x8000U
#define SW3 0x7fffU


#ifndef USE_DEBUG

#ifdef PLAYTIME
    if (USEX(SCI_DECODE_TIME) > PLAYTIME && cachedSpiState != spi_zeros) {
	USEX(SCI_AICTRL0)++;
	USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
	cachedSpiState = spi_zeros;
    }
#endif/*PLAYTIME*/


    if (cachedSpiState == spi_file) {
	/* extra byte read if odd file size */
	spiSize = (fileSize+1)/2; /* in words! */
	//spiAddr = (fileStart<<9);
	spiCnt = 0;

	/* Start one big read from the beginning of the file. */
	USEX(GPIO_ODATA) &= ~SPI_xCS;
	SpiSendReceiveSpi((SPI_READ<<8), 8);
	SpiSendReceiveSpi(fileStart<<1, 16);
	SpiSendReceiveSpi(0, 8);
	cachedSpiState = spi_data;

    } else if (cachedSpiState == spi_data) {
	while (StreamDiff() < STREAM_BUFFER_SZ-40 && spiSize > 0) {
	    register int i, j = 32;
	    register u_int16 *wr = (void *)stream_wr_pointer;
#if 0
	    puthex(spiAddr>>16);
	    puthex(spiAddr);
	    putch(' ');
#endif
	    if (j > spiSize) {
		j = spiSize;
	    }
	    for (i=0;i<j;i++) {
		*wr++ = SpiSendReceiveSpi(0,16);
#if 0
		spiAddr += 2;
		if (i==0) {
		    puthex(wr[-1]);
		}
#endif
		if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
		    wr = stream_buffer;
	    }
	    stream_wr_pointer = (void *)wr;
	    spiSize -= j;
#if 0
	    putch(' ');
	    puthex(spiSize);
	    putch('l');
	    putch('\n');
#endif

	    /* watchdog clearing here doesn't work with midi */
	}

#ifdef ENABLE_WATCHDOG
	USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
	if (spiSize == 0) {
end_of_file:
	    USEX(GPIO_ODATA) |= SPI_xCS;
	    spiCnt = 0;
	    //USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV); not yet!
#if 0
	    putch('e');
	    putch('o');
	    putch('f');
	    putch('\n');
#endif

#ifdef SCI_UI
	    {
		register int m = (USEX(SCI_AICTRL3)& CTRL3_PLAY_MODE_MASK);
		cachedSpiState = spi_zeros;
		if (m == CTRL3_LOOP_SONG) {
		    /* restart the same file immediately */
		    cachedSpiState = spi_file;
#ifdef UART_UI
		    putch('l');//USEX(UART_DATA) = 'l';
#endif
#ifdef UART_FIFO
		    putch('l');//USEX(UART_DATA) = 'l';
		    if (uiLoops > 0) {
			uiLoops--;
			if (uiLoops == 0) {
			    uiLoops = -1; //back to infinite loops
			    USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
			}
		    }
#endif
		} else if (m == CTRL3_PAUSE_AFTER_PLAY) {
		    /* pause after play
		       Note: should be after the zeros have been sent! */
		    USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON | CTRL3_AT_END;
		} else {
		    USEX(SCI_AICTRL0) += 1; /* normal play: next file */
		}
	    }
#else/*SCI_UI*/
	    {
		register int m = (USEX(SCI_AICTRL3)& CTRL3_PLAY_MODE_MASK);
		if (m == CTRL3_LOOP_SONG) {
		    /* restart the same file immediately */
		    cachedSpiState = spi_file;
		} else if (m == CTRL3_PAUSE_AFTER_PLAY) {
		    /* goto pause mode, so next and prev will work to
		       jump to next/prev song.
		       Note: should be after the zeros have been sent!
		    */
		    oldSpiState = spi_zeros;
		    cachedSpiState = spi_pause;
		} else {
		    USEX(SCI_AICTRL0) += 1; /* normal play: next file */
		    cachedSpiState = spi_zeros;
		}
	    }
#endif/*elseSCI_UI*/
	}
    } else if (cachedSpiState == spi_zeros) {
	if (StreamDiff() < STREAM_BUFFER_SZ-40) {
	    register int i, z = 0;
	    register u_int16 *wr = (void *)stream_wr_pointer;
	    for (i=0;i<32;i++) {
		*wr = z;
		if (wr >= &stream_buffer[STREAM_BUFFER_SZ-1]) {
		    wr = stream_buffer;
		} else {
		    wr++;
		}
		++spiCnt;
		stream_wr_pointer = (void *)wr;
	    }
#if 0
	    puthex(spiCnt);
	    putch(' ');
	    putch('c');
	    putch('\n');
#endif
	    /* watchdog clearing here doesn't work with midi */
	}
	if (spiCnt >= 256+2*STREAM_BUFFER_SZ) {
//goreset:
	    USEX(GPIO_ODATA) |= SPI_xCS;
	    USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);

#if defined(UART_UI) || defined(UART_FIFO)
	    putstrp("\pdone\n\n");
#endif
	    MyReset();
	}
#ifdef ENABLE_WATCHDOG
	goto wdogack;//USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
    } else {
#ifdef ENABLE_WATCHDOG
wdogack:
	USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
    }


#ifdef SCI_UI
//    USEX(SER_DREQ) = 0;
#else

#ifndef NO_DREQ_LED
    /* 15 words */
    {
	register int t = 0;

	if ((ui.Timer & 0x18) == 0) {
	    /* DREQ indicates random play status */
#ifdef NO_RANDOM
	    t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
#else
	    t = USEX(SCI_AICTRL3);  /* lowest bit is global random play */
#ifdef LOUDNESS_ALWAYS_ON_OFF
	    /* indicate random play */
#else /*LOUDNESS_ALWAYS_ON_OFF*/

	    if (cachedSpiState
#ifdef SW3_RANDOM_PLAY
		==
#else
		!=
#endif
		spi_pause)
		t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
#endif /*else LOUDNESS_ALWAYS_ON_OFF*/
#endif /*else NO_RANDOM*/
	}
	USEX(SER_DREQ) = t;
    }
#endif/*!NO_DREQ_LED*/
#endif/*!SCI_UI*/
#endif/*!USE_DEBUG*/

    spiState = cachedSpiState;
#ifdef KEYTIMER
#ifdef BATTCHECK
    {
	register s_int16 batt = USEX(MODU_DATA);
	if (ui.KeyTimer >= 0 ||
	    ((ui.KeyTimer & 8) == 0 &&
	     (batt < 0x240 || ui.KeyTimer > (batt>>7) - 60)))
	    USEX(UART_DATA) = 0x00;
    }
#else
    if (ui.KeyTimer >= 0) {
	USEX(UART_DATA) = 0x00;
    }
#endif
#endif/*KEYTIMER*/
}



auto void MyMain(register s_int16 __a1 fullReset) {
#ifdef BATTCHECK
    USEX(MODU_DIV) = 0x0030;
#endif/*BATTCHECK*/

#if defined(SCI_UI) && defined(ENABLE_WATCHDOG)
    USEX(WDOG_CONFIG) = 1000;
    USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif

#if defined(SCI_UI)
    USEX(SCI_AICTRL2) = 0;
#endif
#if 0 && defined(VS1103)
    USEX(SCI_MODE) = (1<<SCIMB_SDI_NEWMODE)
	+ (0<<SCIMB_CONF); /*0=SDIMIDI,SCIWAV 1=SDIWAV,SCIMIDI*/
#endif

    USEX(SCI_MODE) &= ~(1<<SCIMB_OUT_OF_WAV);
#if 0 //For vs1103 greeting card player
    USEX(SCI_MODE) |= (1<<SCIMB_DIFFERENTIAL);
#endif
#ifdef USE_BUTTON
    USEX(GPIO_DDR)   = SPI_xCS|SPI_xCS2;
    USEX(GPIO_ODATA) = SPI_xCS|SPI_xCS2;
#else
    USEX(GPIO_DDR)   = SPI_xCS|SPI_CLK|SPI_xCS2;
    USEX(GPIO_ODATA) = SPI_xCS|SPI_xCS2;
#endif

#if 1
    /* 'wake up' spi memory */
//    USEX(GPIO_ODATA) &= ~SPI_xCS;
//    SpiSendReceiveSpi((SPI_READ<<8), 8);
//    SpiSendReceiveSpi(sector<<1, 16);
    SpiSendReceiveSpi(0, 8);
//    USEX(GPIO_ODATA) |= SPI_xCS;
#endif


    PatchInitHardware(); /* does not clear user area */
#if 1 /*VERSION*/
    /*
      v0.6 print version string
    */
    putword(('S'<<8)|'P');
    putword(('0'<<8)|'.');
    putword(('6'<<8)|'0');
    putch('\n');
#endif
#ifdef UART_FIFO
    UartInit();
#endif

#ifdef UART_UI
    //putch('R');
    //putch('\n');
#endif

    InitCommon();
    //AudioInit(); //rather use..
    //AudioSet(2, 44100U);

#if defined(USE_BUTTON)
    applAddr = ApplAddr;
#endif/*USE_BUTTON*/

    /* Allow DREQ and enable interrupts, VERY IMPORTANT! */
    USEX(INT_ENABLE) = 0;
//    USEX(INT_GLOB_ENA) = 0;
    USEX(INT_GLOB_ENA) = 0;
//    USEX(INT_GLOB_ENA) = 0;

    AudioSet(2, 44100U);
    InitReadFrame();

    /* determine number of playable files */
    USEX(SCI_AICTRL2) = OpenFile(0xffffU);
#if defined(UART_UI) || defined(UART_FIFO)
    putstrp("\pfiles");
    putword(USEX(SCI_AICTRL2));
    putch('\n');
#endif
 again:
#if !defined(NO_RANDOM) /* no random play feature */
    if ((USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY)) {
	/* random play (not shuffle!) */
	USEX(SCI_AICTRL0) = (myrand() % USEX(SCI_AICTRL2));
    }
#endif
    if (OpenFile(USEX(SCI_AICTRL0)
#ifdef SCI_UI
		 &= 0x7fffU //max 32768 files in SCI-controlled version
#endif
	) < 0) {

#if defined(SCI_UI)
	if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
	    CTRL3_PAUSE_BEFORE_PLAY) {
	    /* Mark file ready and go to pause state in UserHook */
	    USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
	}
	/*NEW: set FILE_READY always */
	USEX(SCI_AICTRL3) = CTRL3_FILE_READY | 
	    (USEX(SCI_AICTRL3) & ~CTRL3_AT_END);
#endif


#if defined(PAUSE_AT_POWERON) /*pause at power-on, then use pause after play*/
#ifdef SCI_UI
	if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
	    CTRL3_PAUSE_BEFORE_PLAY) {
	    USEX(SCI_AICTRL3) = CTRL3_PAUSE_AFTER_PLAY | CTRL3_PAUSE_ON;
	}
	spiState = spi_file;

#else/*SCI_UI*/
	if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
	    CTRL3_PAUSE_BEFORE_PLAY) {
	    USEX(SCI_AICTRL3) = CTRL3_PAUSE_AFTER_PLAY
#ifdef SCI_UI
		| CTRL3_PAUSE_ON
#endif
		;
	    oldSpiState = spi_file;
	    spiState = spi_pause;
	    stream_save = stream_wr_pointer;
	} else {
	    spiState = spi_file;
	}
#endif/*elseSCI_UI*/
#else /*PAUSE_AT_POWERON*/
	spiState = spi_file;
#endif /*else PAUSE_AT_POWERON*/

    } else {
    gofirst:
	USEX(SCI_AICTRL0) = 0; /* file not found, jump to first file */
	goto again;
    }

#if !defined(SCI_UI) && defined(WAIT_FOR_BUTTON)
    if ((USEX(SCI_AICTRL3)&CTRL3_PLAY_MODE_MASK) == CTRL3_PAUSE_BEFORE_PLAY) {
	/* Only once per boot */
	USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
	while (1) {
	    ui.Data = 0;
	    /* toggle MOSI to clock SDI */
	    USEX(INT_ENABLE) = ENABLE_SCI_SDI_DAC_RX;
	    ClockSdi();
	    if (ui.Data == 0) {
		USEX(GPIO_ODATA) &= ~SPI_xCS;
		ClockSdi();
		USEX(GPIO_ODATA) |= SPI_xCS;
		ui.Data >>= 1;/* 0xffff or 0x8000 ? -> 0x7fff 0x4000 */
	    }
	    USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
#ifdef ENABLE_WATCHDOG
	    USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
	    if (ui.Data) {
		ui.Key = ui.Data;
		ui.KeyTime = -32767;
		break;
	    }
	    //Halt();//Sleep();
	}
    }
#endif

#if !defined(SCI_UI) && defined(UART_UI) && defined(WAIT_FOR_UART_BYTE)
    /*TODO: lower clock */
    {
	register u_int16 old = USEX(SCI_CLOCKF);
	USEX(SCI_CLOCKF) = 0;
	SetClockFromSpi();

	/*rescue rx if framing error happened.*/
	USEX(UART_DATA);
	USEX(INT_ENABLE) = 0;
	if (1) {
	    register u_int16 d = 0x7f7f;//USEX(UART_DIV);
	    register int i;
	    for (i=0;i<256;i++)
		USEX(UART_DIV) = d;
	}
	USEX(SER_DREQ) = 0;
	while (1) {
	    USEX(SER_DREQ) ^= 1;
#ifdef ENABLE_WATCHDOG
	    USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
	    if (USEX(UART_STATUS) & UART_ST_RXFULL) {
		USEX(UART_DATA);
		break;
	    }
	}
	USEX(SCI_CLOCKF) = old;
	SetClockFromSpi();
    }
#endif


#ifdef USE_BUTTON
    ui.Key = (1<<3);
    ui.KeyTime = -32767;
#endif/*USE_BUTTON*/

#if defined(ENABLE_WATCHDOG) && defined(KEEP_DRIVERS_OFF)
    /*otherwise keep it in powerdown until MMC found*/
    USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1); /* enable ANALOG power */
#endif

    USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
    USEX(SCI_DECODE_TIME) = 0;

#if defined(UART_UI) || defined(UART_FIFO)
    putstrp("\pplay");
    putword(USEX(SCI_AICTRL0));
#ifdef SCI_UI
    putword(fileName[0]);
    putword(fileName[1]);
    putword(fileName[2]);
    putword(fileName[3]);
    putword(fileName[4]);
    putch15(fileName[5]);
    putch('\n');
#endif/*SCI_UI*/
#endif
    /*A modified tradpatch for VS1103B */
    {
#define MIXER_OUT_SAMPLES 34
	auto s_int16 DecodeHeader(register __d u_int32 head);
	struct Wav {
	    s_int16 channels;
	    u_int16 sampleRate;
	    s_int16 ima;
	    s_int16 bitsPerSample;
	    s_int16 samplesPerBlock;
	    u_int16 imaSync;
	    u_int32 samplesLeft;
	    u_int16 flags;
	};
	extern struct Wav wav;
	extern s_int16 mixerOut[3][2*MIXER_OUT_SAMPLES];

	register u_int32 head = 0;

#if 0
	/* test DAC latency */
	SetHardware(2, 44100U);
	Disable();
	USEX(SCI_STATUS) &= ~0xf;
	while (1) {
	    while (!(USEX(UART_STATUS)& UART_ST_RXFULL))
		;
	    USEX(UART_DATA);
	    USEX(UART_DATA) = 0x12;
	    USEX(DAC_LEFT) = USEX(DAC_RIGHT) = 0x7fff;
	    while (!(USEX(UART_STATUS)& UART_ST_RXFULL))
		;
	    USEX(UART_DATA);
	    USEX(UART_DATA) = 0x23;
	    USEX(DAC_LEFT) = USEX(DAC_RIGHT) = 0x8000;
	}
#endif


	skipMixer = 1;
	while (1) {
	    if (StreamDiff()) {
		head <<= 8;
		head |= SdiGetC();
//puthex(head);
//putch('\n');


		if (head == MAKE_ID('R','I','F','F')) {
		    ResetAdpcm();
#if 1
		    /* Move "RIFF" back to input buffer */
		    if ((stream_rd_pointer -= 2) < stream_buffer)
			stream_rd_pointer += STREAM_BUFFER_SZ;
#else
		    /* we know we always start from the beginning
		       -- except loop! */
		    stream_rd_pointer = stream_buffer;
#endif
		    adpcmSource = 0;	/* ADPCM from SDI */
		    while (1) {
			GetAdpcmSamples(NULL, mixerOut[0], 32);
			if (wav.sampleRate && wav.sampleRate != hwSampleRate) {
			    SetHardware(wav.channels, wav.sampleRate);
			}
			WmaStereoCopy(mixerOut[0], 32);
		    }
		} else {
//		    USEX(SCI_DECODE_TIME) = 0;
		    DecodeHeader(head);
		}
	    } else {
		/*required for key handling (pause)*/
		memset(pcm_sample, 0, AUDIOBUFSIZE);
		wav_write();
	    }
	}
    }
}
