/*
  VS1103b Simple SPI Flash Recorder / Player
  ==========================================
  This source code package is provided as-is and without warranty.
  (C)2008 VLSI Solution Oy

  This code is licensed for free use with any product using VS1103b.

  Some features can be turned on and off with preprocessor definitions
  in Makefile and in this file. Features that are enabled have an
  effect on the code size, and the resulting program may not fit in
  the available code space. In this case the linker will give an error.
*/

/*
  Note on MOSI/DREQ:

  DREQ is changed in normal firmware both in SDI interrupts when the
  stream buffer becomes too full, and whenever the stream buffer becomes
  empty enough.

  In this version SDI interrupt is not enabled, so it does not touch DREQ.
  So, we can use DREQ ourselves as we wish and access to it does not need to
  be protected by interrupt Disable()/Enable(). However, the normal decoding
  will still keep raising DREQ when it sees that there is space for new data.

 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "module.h"
#include "h1103/ima.h"
#include "h1103/imalib.h"
#include "spi.h"
#undef puts
#undef printf
#undef putchar
#undef fputs


#ifndef SpiPrivDelay
auto void SpiPrivDelay(void) {
    register int i;
    for (i=0;i<10;i++) {
	USEX(SER_DREQ);
    }
}
#endif

#if 1
auto u_int16 SpiSendReceiveMmc(register __a0 u_int16 dataTopAligned,
			       register __a1 bits);
#define SpiPrivSendReceive SpiSendReceiveMmc
#else
auto u_int16 SpiPrivSendReceive(register __a0 u_int16 dat,
				register __a1 s_int16 bits) {
  register u_int16 odata = 0;
  while (bits--) {
    USEX(SER_DREQ) = (dat>>bits);
#ifndef WORK_FASTER
    SpiPrivDelay();
#endif
    USEX(GPIO_ODATA) |= SPI_CLK;
#ifndef WORK_FASTER
    SpiPrivDelay();
#endif
    odata <<= 1;
#if SPI_MISO == 4
    odata |= (USEX(GPIO_IDATA) >> 2) & 1;
#else
    odata |= (USEX(GPIO_IDATA) & SPI_MISO) ? 1 : 0;
#endif
    USEX(GPIO_ODATA) &= ~SPI_CLK;
  }
  return odata;
}
#endif

auto void SpiPrivInit(void) {
  USEX(GPIO_ODATA) = SPI_xCS;
  USEX(GPIO_DDR) = SPI_xCS|SPI_CLK;
  USEX(GPIO_ODATA) = SPI_xCS;
}

auto u_int16 SpiPrivStatus(void) {
  register u_int16 dat;

  USEX(GPIO_ODATA) |= SPI_xCS;
  SpiPrivDelay();
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  SpiPrivSendReceive(SPI_RDSR, 8);
  dat = SpiPrivSendReceive(0, 8);
  SpiPrivDelay();
  USEX(GPIO_ODATA) |= SPI_xCS;
  return dat;
}



void SingleCycle8BitCommand(u_int16 cmd) {
  while (SpiPrivStatus() & 1)
    SpiPrivDelay();
  USEX(GPIO_ODATA) |= SPI_xCS;
  SpiPrivDelay();
  USEX(GPIO_ODATA) &= ~SPI_xCS;
  SpiPrivSendReceive(cmd, 8);
  SpiPrivDelay();
  USEX(GPIO_ODATA) |= SPI_xCS;
}
